//
//  ViewController.swift
//  KeepScrollPositionDemo
//
//  Created by Steffen Neubauer on 22/04/16.
//  Copyright © 2016 Steffen Neubauer. All rights reserved.
//

import UIKit

class DemoViewController: TLTableViewController {

    let LoadingCellIdentifierBefore = "LoadingBefore"
    let LoadingCellIdentifierAfter = "LoadingAfter"
    let ExampleCellReuseIdenfier = "ExampleCell"

    var isLoadingInitially = true
    var counter = 0
    var items: [TLIndexPathItem] = [] {
        didSet {
            self.indexPathController.items = items
        }
    }

    override func viewDidLoad() {
        super.viewDidLoad()

        self.tableView.registerClass(UITableViewCell.self, forCellReuseIdentifier: LoadingCellIdentifierAfter)
        self.tableView.registerClass(UITableViewCell.self, forCellReuseIdentifier: LoadingCellIdentifierBefore)
        self.tableView.registerClass(UITableViewCell.self, forCellReuseIdentifier: ExampleCellReuseIdenfier)

        self.items = self.createRandomColors()
    }
    
    override func tableView(tableView: UITableView, didSelectRowAtIndexPath indexPath: NSIndexPath) {
        
        self.navigationController!.pushViewController(DemoViewController(), animated: true)
    }
    
    override func controller(controller: TLIndexPathController, willUpdateDataModel oldDataModel: TLIndexPathDataModel?, withDataModel updatedDataModel: TLIndexPathDataModel?) -> TLIndexPathDataModel? {

        if updatedDataModel!.items.count == 0 {
            return nil
        }
        
        let loadingBefore = TLIndexPathItem(
            identifier: LoadingCellIdentifierBefore,
            sectionName: nil,
            cellIdentifier: LoadingCellIdentifierBefore,
            data: nil)
        let loadingAfter = TLIndexPathItem(
            identifier: LoadingCellIdentifierAfter,
            sectionName: nil,
            cellIdentifier: LoadingCellIdentifierAfter,
            data: nil)

        return TLIndexPathDataModel(items: [loadingBefore] + updatedDataModel!.items + [loadingAfter])
    }
    
    override func tableView(tableView: UITableView, willDisplayCell cell: UITableViewCell, forRowAtIndexPath indexPath: NSIndexPath) {
        
        let item = self.itemAtIndexPath(indexPath)

        if item.cellIdentifier == LoadingCellIdentifierBefore {
            delay(0.3) {
                self.items = self.createRandomColors() + self.items
            }
        } else if item.cellIdentifier == LoadingCellIdentifierAfter {
            delay(0.3) {
                self.items = self.items + self.createRandomColors()
            }
        }
    }
    
    override func tableView(tableView: UITableView!, configureCell cell: UITableViewCell!, atIndexPath indexPath: NSIndexPath!) {
        
        let item = self.itemAtIndexPath(indexPath)
        
        if let color = item.data as? UIColor {
            cell.contentView.backgroundColor = color
        } else {
            cell.textLabel?.text = "Loading..."
            cell.textLabel?.textAlignment = .Center
        }
    }
    
    override func tableView(tableView: UITableView, heightForRowAtIndexPath indexPath: NSIndexPath) -> CGFloat {
        return 140
    }
    
    override func controller(controller: TLIndexPathController, didUpdateDataModel updates: TLIndexPathUpdates) {
        
        let (referenceItem, offset) = self.findReferenceItem(updates)
        
        UIView.performWithoutAnimation {
            if self.isViewLoaded() && self.view.window != nil {
                self.tableView.beginUpdates()
                updates.performBatchUpdatesOnTableView(self.tableView, withRowAnimation: .Fade)

                if let referenceItem = referenceItem {
                    self.scrollToItem(referenceItem, withOffset: offset)
                }
                self.tableView.endUpdates()
            } else {
                self.tableView.reloadData()
                self.fixInitialScrollPositionIfNeeded()
            }
        }
    }

    private func fixInitialScrollPositionIfNeeded() {
        if self.isLoadingInitially {
            self.isLoadingInitially = false
            self.tableView.scrollToRowAtIndexPath(
                NSIndexPath(forRow: 1, inSection: 0),
                atScrollPosition: .Top,
                animated: false)
        }
    }

    private func findReferenceItem(updates: TLIndexPathUpdates) -> (AnyObject?, CGFloat) {
        if let referenceItem = getItemBothVisibleAndInUpdates(updates) {
            let indexPath = updates.oldDataModel!.indexPathForItem(referenceItem)!
            let rect = self.tableView.rectForRowAtIndexPath(indexPath)
            let offset = self.tableView.contentOffset.y - rect.origin.y;

            return (referenceItem, offset)
        }

        return (nil, 0)
    }

    private func getItemBothVisibleAndInUpdates(updates: TLIndexPathUpdates) -> AnyObject?  {
        if updates.oldDataModel == nil {
            return nil
        }
        
        let itemBothVisibleAndInUpdates = (self.tableView.indexPathsForVisibleRows ?? [])
            .map {
                updates.oldDataModel!.itemAtIndexPath($0) as! TLIndexPathItem
            }.filter {
                return $0.data is UIColor
            }.filter {
                return updates.updatedDataModel!.containsItem($0)
            }.first
        
        return itemBothVisibleAndInUpdates
    }

    private func scrollToItem(item: AnyObject, withOffset offset: CGFloat) {
        let indexPath = self.indexPathController.dataModel!.indexPathForItem(item)!
        
        let rect = self.tableView.rectForRowAtIndexPath(indexPath)
        var contentOffset = self.tableView.contentOffset
        contentOffset.y = rect.origin.y + offset
        self.tableView.contentOffset = contentOffset
    }

    private func itemAtIndexPath(indexPath: NSIndexPath) -> TLIndexPathItem {
        return self.indexPathController.dataModel!.itemAtIndexPath(indexPath)! as! TLIndexPathItem
    }

    private func createRandomColors() -> [TLIndexPathItem] {
        return [
            UIColor.redColor(),
            UIColor.blackColor(),
            UIColor.blueColor(),
            UIColor.greenColor(),
            UIColor.cyanColor(),
            UIColor.darkGrayColor(),
            UIColor.brownColor(),
            UIColor.orangeColor(),
            UIColor.purpleColor(),
        ].shuffle().map {
            self.counter += 1
            
            return TLIndexPathItem(
                identifier: self.counter,
                sectionName: nil,
                cellIdentifier: ExampleCellReuseIdenfier,
                data: $0)
        }
    }

}

func delay(delay:Double, closure:()->()) {
    dispatch_after(
        dispatch_time(
            DISPATCH_TIME_NOW,
            Int64(delay * Double(NSEC_PER_SEC))
        ),
        dispatch_get_main_queue(), closure)
}

extension CollectionType {
    /// Return a copy of `self` with its elements shuffled
    func shuffle() -> [Generator.Element] {
        var list = Array(self)
        list.shuffleInPlace()
        return list
    }
}

extension MutableCollectionType where Index == Int {
    /// Shuffle the elements of `self` in-place.
    mutating func shuffleInPlace() {
        // empty and single-element collections don't shuffle
        if count < 2 { return }
        
        for i in 0..<count - 1 {
            let j = Int(arc4random_uniform(UInt32(count - i))) + i
            guard i != j else { continue }
            swap(&self[i], &self[j])
        }
    }
}
